/*
 *   Copyright (C) 2002-2004 by Jonathan Naylor G4KLX
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "JT44App.h"
#include "JT44Defs.h"

#include <wx/config.h>
#include <wx/filename.h>
#include <wx/ffile.h>

IMPLEMENT_APP(CJT44App)

#include "common/SoundCard.h"
#include "common/Exception.h"
#include "common/Log.h"

const wxString SECTION_GENERAL  = wxT("General");
const wxString SECTION_TRANSMIT = wxT("Transmit");
const wxString SECTION_RECEIVE  = wxT("Receive");
const wxString SECTION_SOUND    = wxT("Sound");
const wxString SECTION_PTT      = wxT("PTT");
const wxString SECTION_PATHS    = wxT("Paths");
const wxString SECTION_MESSAGES = wxT("Messages");

const wxString KEY_CALLSIGN     = wxT("callsign");
const wxString KEY_LOCATOR      = wxT("locator");
const wxString KEY_BAND         = wxT("band");
const wxString KEY_FIRST        = wxT("first");
const wxString KEY_MIN_SYNC     = wxT("min_sync");
const wxString KEY_TYPE         = wxT("type");
const wxString KEY_FILENAME     = wxT("filename");
const wxString KEY_DEVICE       = wxT("device");
const wxString KEY_AUDIO        = wxT("audio");
const wxString KEY_TEXT         = wxT("text");
const wxString KEY_MESSAGE      = wxT("message_");

const wxString DEFAULT_CALLSIGN = wxT("G4KLX");
const wxString DEFAULT_LOCATOR  = wxT("IO93FB");
const long     DEFAULT_BAND     = 144L;

const bool     DEFAULT_TX_ENABLED = false;
const bool     DEFAULT_TX_FIRST   = true;

const long     DEFAULT_MIN_SYNC = 1L;

const wxString DEFAULT_PTT_TYPE   = wxT("none");
const wxString DEFAULT_PTT_DEVICE = wxEmptyString;

#ifdef __WINDOWS__
const wxString DEFAULT_SOUND_FILENAME = wxT("Wave Mapper");
#else
const wxString DEFAULT_SOUND_FILENAME = wxT("/dev/dsp");
#endif

const wxString DEFAULT_MESSAGE_0 = wxT("%T %M ");
const wxString DEFAULT_MESSAGE_1 = wxT("%T %M %R%R%R%R%R%R%R%R ");
const wxString DEFAULT_MESSAGE_2 = wxT("R%R");
const wxString DEFAULT_MESSAGE_3 = wxT("R");
const wxString DEFAULT_MESSAGE_4 = wxT("73");
const wxString DEFAULT_MESSAGE_5 = wxT("CQ %M ");


CJT44App::CJT44App() :
wxApp(),
m_frame(NULL),
m_controller(NULL),
m_messages(NULL),
m_file(NULL),
m_saving(false),
m_recording(false),
m_txEnable(DEFAULT_TX_ENABLED)
{
}

CJT44App::~CJT44App()
{
}

bool CJT44App::OnInit()
{
	SetVendorName(JT44_VENDOR_NAME);
	SetAppName(JT44_APPL_NAME);

	wxConfigBase* config = wxConfigBase::Get();
	config->SetRecordDefaults();

	wxString pathAudio, pathText;
	getPaths(pathAudio, pathText);

	wxFileName fileName(pathText, JT44_LOGFILE_NAME, JT44_LOGFILE_EXT);

	wxLog* log = new CLog(fileName.GetFullPath());
	wxLog::SetActiveTarget(log);

	createController();
	::wxSleep(1);

	createFileReader();

	readRegistry();

	m_frame = new CJT44Frame();
	m_frame->Show();

	SetTopWindow(m_frame);

	return true;
}

int CJT44App::OnExit()
{
	wxASSERT(m_controller != NULL);
	wxASSERT(m_file != NULL);

	m_controller->Delete();
	m_file->kill();

	delete wxConfigBase::Set(NULL);

	return 0;
}

#ifdef __WXDEBUG__
void CJT44App::OnAssert(const wxChar* file, int line, const wxChar* cond, const wxChar* msg)
{
	::wxLogFatalError(wxT("Assertion failed on line %d in file %s: %s %s"), line, file, cond, msg);
}
#endif

void CJT44App::receiveMessage(CJT44Message* message, EWho who) const
{
	wxASSERT(m_frame != NULL);
	wxASSERT(message != NULL);

	m_frame->showMessage(message, who);
}

void CJT44App::logMessage(const CJT44Message& message) const
{
	if (m_saving) {
		wxString     id = message.getId();
		int     quality = message.getQuality();
		double       Dt = message.getDt();
		int          DF = message.getDF();
		double strength = message.getStrength();

		wxString   text = message.getText();
		wxString oeText = message.getOddEvenText();
		wxString l4Text = message.getLast4Text();

		wxString pathAudio, pathText;
		getPaths(pathAudio, pathText);

		wxFileName fileName(pathText, JT44_MESSAGES_NAME, JT44_MESSAGES_EXT);

		wxFFile* file = new wxFFile(fileName.GetFullPath(), "a");

		if (!file->IsOpened()) {
			wxString text;
			text.Printf(wxT("Cannot open save text file: %s"), fileName.GetFullPath().c_str());
			error(text);

			delete file;
			return;
		}

		wxString line;
		line.Printf(wxT("%s  %d  %.0f  %.1f  %d  %s  %s  %s\n"), id.c_str(), quality, strength - 0.5, Dt, DF, text.c_str(), oeText.c_str(), l4Text.c_str());

		file->Write(line);
		file->Close();

		delete file;
	}
}

void CJT44App::showLevels(CJT44Levels* levels, EWho who) const
{
	wxASSERT(levels != NULL);
	wxASSERT(m_frame != NULL);

	m_frame->showLevels(levels, who);
}

void CJT44App::showAudio(double audio, EWho who) const
{
	wxASSERT(m_frame != NULL);

	m_frame->showAudio(audio, who);
}

void CJT44App::error(const wxString& text) const
{
	wxASSERT(m_frame != NULL);

	m_frame->error(text);
}

void CJT44App::getPersonal(wxString& callsign, wxString& locator) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString callsignKey = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_CALLSIGN;
	wxString locatorKey  = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_LOCATOR;

	profile->Read(callsignKey, &callsign, DEFAULT_CALLSIGN);
	profile->Read(locatorKey,  &locator,  DEFAULT_LOCATOR);
}

void CJT44App::setPersonal(const wxString& callsign, const wxString& locator) const
{
	CText::setLocalCallsign(callsign);
	CText::setLocalLocator(locator);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString callsignKey = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_CALLSIGN;
	wxString locatorKey  = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_LOCATOR;

	profile->Write(callsignKey, callsign);
	profile->Write(locatorKey,  locator);
	profile->Flush();
}

int CJT44App::getBand() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_BAND;

	long band;
	profile->Read(key, &band, DEFAULT_BAND);

	return int(band);
}

void CJT44App::setBand(int band) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_GENERAL + wxT("/") + KEY_BAND;

	profile->Write(key, long(band));
	profile->Flush();
}

void CJT44App::setRemote(const wxString& callsign, const wxString& report)
{
	if (callsign.IsEmpty())
		CText::setRemoteCallsign(wxEmptyString);
	else
		CText::setRemoteCallsign(callsign);

	if (report.IsEmpty())
		CText::setLocalReport(wxEmptyString);
	else
		CText::setLocalReport(report);
}

bool CJT44App::getTXFirst() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_TRANSMIT + wxT("/") + KEY_FIRST;

	bool txFirst;
	profile->Read(key, &txFirst, DEFAULT_TX_FIRST);

	return txFirst;
}

void CJT44App::setTXFirst(bool txFirst) const
{
	wxASSERT(m_controller != NULL);

	m_controller->setTXFirst(txFirst);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_TRANSMIT + wxT("/") + KEY_FIRST;

	profile->Write(key, txFirst);
	profile->Flush();
}

bool CJT44App::getTXEnable() const
{
	return m_txEnable;
}

void CJT44App::setTXEnable(bool txEnable)
{
	wxASSERT(m_controller != NULL);

	m_controller->setTXEnable(txEnable);

	m_txEnable = txEnable;
}

int CJT44App::getMinSync() const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_RECEIVE + wxT("/") + KEY_MIN_SYNC;

	long minSync;
	profile->Read(key, &minSync, DEFAULT_MIN_SYNC);

	return int(minSync);
}

void CJT44App::setMinSync(int minSync) const
{
	wxASSERT(m_controller != NULL);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString key = wxT("/") + SECTION_RECEIVE + wxT("/") + KEY_MIN_SYNC;

	profile->Write(key, minSync);
	profile->Flush();
}

wxString CJT44App::getReplacedMessage(int n) const
{
	wxASSERT(n >= 0 && n < JT44_MAX_MESSAGES);
	wxASSERT(m_messages[n] != NULL);

	return m_messages[n]->getReplacedText(JT44_MESSAGE_LENGTH);
}

wxString CJT44App::getOriginalMessage(int n) const
{
	wxASSERT(n >= 0 && n < JT44_MAX_MESSAGES);
	wxASSERT(m_messages[n] != NULL);

	return m_messages[n]->getOriginalText();
}

void CJT44App::getMessage(int n, wxString& message) const
{
	wxASSERT(n >= 0 && n < JT44_MAX_MESSAGES);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString number;
	number.Printf(wxT("%d"), n);

	wxString key = wxT("/") + SECTION_MESSAGES + wxT("/") + KEY_MESSAGE + number;
	wxString defMessage;

	switch (n) {
		case 0:
			defMessage = DEFAULT_MESSAGE_0;
			break;
		case 1:
			defMessage = DEFAULT_MESSAGE_1;
			break;
		case 2:
			defMessage = DEFAULT_MESSAGE_2;
			break;
		case 3:
			defMessage = DEFAULT_MESSAGE_3;
			break;
		case 4:
			defMessage = DEFAULT_MESSAGE_4;
			break;
		case 5:
			defMessage = DEFAULT_MESSAGE_5;
			break;
		default:
			defMessage = wxEmptyString;
			break;
	}

	profile->Read(key, &message, defMessage);
}

void CJT44App::setMessage(int n, const wxString& message)
{
	wxASSERT(n >= 0 && n < JT44_MAX_MESSAGES);

	if (m_messages[n] != NULL)
		delete m_messages[n];

	m_messages[n] = new CText(message);

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString number;
	number.Printf(wxT("%d"), n);

	wxString key = wxT("/") + SECTION_MESSAGES + wxT("/") + KEY_MESSAGE + number;

	profile->Write(key, message);
	profile->Flush();
}

void CJT44App::getPTT(wxString& type, wxString& device) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString typeKey   = wxT("/") + SECTION_PTT + wxT("/") + KEY_TYPE;
	wxString deviceKey = wxT("/") + SECTION_PTT + wxT("/") + KEY_DEVICE;

	profile->Read(typeKey,   &type,   DEFAULT_PTT_TYPE);
	profile->Read(deviceKey, &device, DEFAULT_PTT_DEVICE);
}

void CJT44App::setPTT(const wxString& type, const wxString& device) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString typeKey   = wxT("/") + SECTION_PTT + wxT("/") + KEY_TYPE;
	wxString deviceKey = wxT("/") + SECTION_PTT + wxT("/") + KEY_DEVICE;

	profile->Write(typeKey,   type);
	profile->Write(deviceKey, device);
	profile->Flush();
}

void CJT44App::getPaths(wxString& pathAudio, wxString& pathText) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString audioKey = wxT("/") + SECTION_PATHS + wxT("/") + KEY_AUDIO;
	wxString textKey  = wxT("/") + SECTION_PATHS + wxT("/") + KEY_TEXT;

	wxString homeDir = wxFileName::GetHomeDir();

	profile->Read(audioKey, &pathAudio, homeDir);
	profile->Read(textKey,  &pathText,  homeDir);
}

void CJT44App::setPaths(const wxString& pathAudio, const wxString& pathText) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString audioKey = wxT("/") + SECTION_PATHS + wxT("/") + KEY_AUDIO;
	wxString textKey  = wxT("/") + SECTION_PATHS + wxT("/") + KEY_TEXT;

	profile->Write(audioKey, pathAudio);
	profile->Write(textKey,  pathText);
	profile->Flush();
}

void CJT44App::getSound(wxString& fileName) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString fileNameKey = wxT("/") + SECTION_SOUND + wxT("/") + KEY_FILENAME;

	profile->Read(fileNameKey, &fileName, DEFAULT_SOUND_FILENAME);
}

void CJT44App::setSound(const wxString& fileName) const
{
	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	wxString fileNameKey = wxT("/") + SECTION_SOUND + wxT("/") + KEY_FILENAME;

	profile->Write(fileNameKey, fileName);
	profile->Flush();
}

void CJT44App::setSaving(bool saving)
{
	m_saving = saving;
}

void CJT44App::deleteMessages() const
{
	wxString pathAudio, pathText;
	getPaths(pathAudio, pathText);

	wxFileName fileName(pathText, JT44_MESSAGES_NAME, JT44_MESSAGES_EXT);

	::wxRemoveFile(fileName.GetFullPath());
}

void CJT44App::setRecording(bool recording)
{
	wxASSERT(m_controller != NULL);

	m_recording = recording;

	m_controller->setRXRecording(m_recording);
}

void CJT44App::sendMessage(const wxString& text)
{
	wxASSERT(m_controller != NULL);

	m_controller->sendMessage(text);
}

void CJT44App::readFile(const wxString& fileName)
{
	wxASSERT(m_file != NULL);

	if (m_file->isRunning()) {
		error(wxT("Another WAV file is currently being processed"));
		return;
	}

	try {
		m_file->setFileName(fileName);
		m_file->start();
	}
	catch (CException& ex) {
		error(ex.getMessage());
		return;
	}
}

void CJT44App::createController()
{
	try {
		wxString type, fileName;

		getSound(fileName);
		CSoundCard* soundDev = new CSoundCard(fileName, JT44_SAMPLE_RATE, JT44_SOUNDBUF_LENGTH, JT44_SYMBOL_LENGTH);

		getPTT(type, fileName);
		CPTTPort* pttPort = CPTTPort::createPTTPort(type, fileName);

		m_controller = new CJT44Controller(soundDev, pttPort);
		m_controller->Create();
		m_controller->Run();
	}
	catch (CException& ex) {
		::fprintf(stderr, "JT44: an error has occured when creating the sound device or the PTT port: %s\n", ex.getMessage().c_str());
		throw;
	}
}

void CJT44App::createFileReader()
{
	m_file = new CJT44FileReceive(wxT("File Receive"), RX_FILE);

	m_file->Create();
	m_file->Run();
}

void CJT44App::readRegistry()
{
	wxASSERT(m_controller != NULL);

	wxString callsign, locator;
	getPersonal(callsign, locator);
	CText::setLocalCallsign(callsign);
	CText::setLocalLocator(locator);

	bool txEnable = getTXEnable();
	m_controller->setTXEnable(txEnable);

	bool txFirst = getTXFirst();
	m_controller->setTXFirst(txFirst);

	m_messages = new CText*[JT44_MAX_MESSAGES];

	for (int i = 0; i < JT44_MAX_MESSAGES; i++) {
		wxString message;
		getMessage(i, message);
		m_messages[i] = new CText(message);
	}

	wxConfigBase* profile = wxConfigBase::Get();

	wxASSERT(profile != NULL);

	profile->Flush();
}
